/*
 * Copyright (C) 2007-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

#include <regex.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "ipc.h"

/** \file playlist.c Functions for loading and saving playlists. */

/** Structure used for adding entries to a playlist. */
struct playlist_info {
	/** The name of the playlist. */
	char *name;
	/** The number of entries currently in the playlist. */
	unsigned length;
};
static struct playlist_info current_playlist;

/**
 * Re-insert an audio file into the tree of admissible files.
 *
 * \param aft_row Determines the audio file.
 *
 * \return The return value of score_update().
 */
static int playlist_update_audio_file(const struct osl_row *aft_row)
{
	/* always re-insert to the top of the tree */
	return score_update(aft_row, 0);
}

static int add_playlist_entry(char *path, void *data)
{
	struct playlist_info *playlist = data;
	struct osl_row *aft_row;
	int ret = aft_get_row_of_path(path, &aft_row);

	if (ret < 0) {
		PARA_NOTICE_LOG("%s: %s\n", path, para_strerror(-ret));
		return 1;
	}
	ret = score_add(aft_row, -playlist->length);
	if (ret < 0) {
		PARA_ERROR_LOG("failed to add %s: %s\n", path, para_strerror(-ret));
		return ret;
	}
	playlist->length++;
	return 1;
}

/* returns -E_PLAYLIST_LOADED on _success_ to terminate the loop */
static int load_playlist(struct osl_row *row, void *data)
{
	struct playlist_info *playlist = data;
	struct osl_object playlist_def;
	char *playlist_name;
	int ret;

	ret = pl_get_name_and_def_by_row(row, &playlist_name, &playlist_def);
	if (ret < 0)
		goto err;
	playlist->length = 0;
	ret = for_each_line_ro(playlist_def.data, playlist_def.size,
		add_playlist_entry, playlist);
	osl_close_disk_object(&playlist_def);
	if (ret < 0)
		goto err;
	ret = -E_PLAYLIST_EMPTY;
	if (!playlist->length)
		goto err;
	playlist->name = para_strdup(playlist_name);
	PARA_NOTICE_LOG("loaded playlist %s (%u files)\n", playlist->name,
		playlist->length);
	return -E_PLAYLIST_LOADED;
err:
	if (ret != -E_DUMMY_ROW)
		PARA_NOTICE_LOG("unable to load playlist (%s)\n",
			para_strerror(-ret));
	return 1;
}

static int check_playlist_path(char *path, void *data)
{
	struct para_buffer *pb = data;
	struct osl_row *aft_row;
	int ret = aft_get_row_of_path(path, &aft_row);

	if (ret >= 0)
		return 1;
	return para_printf(pb, "%s: %s\n", path, para_strerror(-ret));
}

static int check_playlist(struct osl_row *row, void *data)
{
	struct para_buffer *pb = data;
	struct osl_object playlist_def;
	char *playlist_name;
	int ret = pl_get_name_and_def_by_row(row, &playlist_name, &playlist_def);

	if (ret < 0)
		return para_printf(pb, "failed to get playlist data: %s\n",
			para_strerror(-ret));
	if (*playlist_name) { /* skip dummy row */
		ret = para_printf(pb, "checking playlist %s...\n", playlist_name);
		if (ret < 0)
			return ret;
		ret = for_each_line_ro(playlist_def.data, playlist_def.size,
			check_playlist_path, pb);
	}
	osl_close_disk_object(&playlist_def);
	return ret;
}

/**
 * Check the playlist table for inconsistencies.
 *
 * \param fd The afs socket.
 * \param query Unused.
 *
 * \return The return value of the underlying call to osl_rbtree_loop().
 */
void playlist_check_callback(int fd, __a_unused const struct osl_object *query)
{
	struct para_buffer pb = {
		.max_size = SHMMAX,
		.private_data = &fd,
		.max_size_handler = pass_buffer_as_shm
	};
	int ret = para_printf(&pb, "checking playlists...\n");

	if (ret < 0)
		return;
	osl_rbtree_loop(playlists_table, BLOBCOL_ID, &pb,
		check_playlist);
	if (pb.offset)
		pass_buffer_as_shm(pb.buf, pb.offset, &fd);
	free(pb.buf);
}

/**
 * Close the current playlist.
 *
 * \sa playlist_open().
 */
void playlist_close(void)
{
	if (!current_playlist.name)
		return;
	free(current_playlist.name);
	current_playlist.name = NULL;
}

/**
 * Open the given playlist.
 *
 * \param name The name of the playlist to open.
 *
 * Files which are listed in the playlist, but not contained in the database
 * are ignored.  This is not considered an error.
 *
 * \return Standard.
 */
int playlist_open(char *name)
{
	struct osl_object obj;
	int ret;
	struct osl_row *row;

	obj.data = name;
	obj.size = strlen(obj.data);
	ret = osl(osl_get_row(playlists_table, BLOBCOL_NAME, &obj, &row));
	if (ret < 0) {
		PARA_NOTICE_LOG("failed to load playlist %s\n", name);
		return ret;
	}
	playlist_close();
	ret = load_playlist(row, &current_playlist);
	return (ret == -E_PLAYLIST_LOADED)? current_playlist.length : ret;
}

static int search_path(char *path, void *data)
{
	if (strcmp(path, data))
		return 1;
	return -E_PATH_FOUND;
}

static int handle_audio_file_event(enum afs_events event, void *data)
{
	int ret, was_admissible = 0, is_admissible;
	struct osl_object playlist_def;
	char *new_path;
	const struct osl_row *row = data;

	if (!current_playlist.name)
		return 1;
	if (event == AUDIO_FILE_RENAME) {
		ret = row_belongs_to_score_table(row, NULL);
		if (ret < 0)
			return ret;
		was_admissible = ret;
	}
	ret = get_audio_file_path_of_row(row, &new_path);
	if (ret < 0)
		return ret;
	ret = pl_get_def_by_name(current_playlist.name, &playlist_def);
	if (ret < 0)
		return ret;
	ret = for_each_line_ro(playlist_def.data, playlist_def.size,
		search_path, new_path);
	osl_close_disk_object(&playlist_def);
	is_admissible = (ret < 0);
	if (was_admissible && is_admissible)
		return 1;
	if (!was_admissible && !is_admissible)
		return 1;
	if (was_admissible && !is_admissible) {
		current_playlist.length--;
		return score_delete(row);
	}
	/* !was_admissible && is_admissible */
	current_playlist.length++;
	return score_add(row, 0); /* play it immediately */
}

/**
 * Handle afs events relevant to playlists.
 *
 * \param event The event type.
 * \param pb Unused.
 * \param data Depends on the event type.
 *
 * \return Standard.
 */
int playlists_event_handler(enum afs_events event,
	__a_unused struct para_buffer *pb, void *data)
{
	int ret;
	struct afsi_change_event_data *aced = data;

	switch(event) {
	case AFSI_CHANGE:
		return playlist_update_audio_file(aced->aft_row);
	case AUDIO_FILE_RENAME:
	case AUDIO_FILE_ADD:
		return handle_audio_file_event(event, data);
	case AUDIO_FILE_REMOVE:
		ret = row_belongs_to_score_table(data, NULL);
		if (ret < 0)
			return ret;
		if (!ret)
			return 1;
		current_playlist.length--;
		return score_delete(data);
	default:
		return 1;
	}
}
